﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.IO;
using System.Runtime.Serialization;
using System.Diagnostics;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Reflection;

namespace CarFire
{
    /// <summary>
    /// A map object represents the map or virtual world where players and other
    /// game entities exist.  The map consists of a grid where each grid space can
    /// contain static scenery and/or game entities which can move and interact
    /// with other game entities.
    /// </summary>
    public class Map
    {
        #region Public Constants

        public const float PixelsToUnitSquares = 64.0f;

        #endregion


        #region Public Types

        /// <summary>
        /// The type of a map helps determine how the map is intended to be used.
        /// </summary>
        public enum Mode
        {
            None,
            Campaign,
            Battle
        }

        /// <summary>
        /// The container class for map metadata.
        /// </summary>
        public class Metadata
        {
            public string Name;
            public Mode Type;
            public string Author;
            public HashSet<int> NumPlayers = new HashSet<int>();
            public string Tileset;
            public int GridWidth;
            public int GridHeight;
        }

        /// <summary>
        /// The container class for information about an entity defined in the map.
        /// </summary>
        public class RawEntity
        {
            public char Id;
            public Point Position;
            public Dictionary<string, string> Attributes = new Dictionary<string, string>();
        }

        #endregion


        #region Public Properties

        /// <summary>
        /// Get the name of the map.
        /// </summary>
        public string Name { get { return mData.Metadata.Name; } }

        /// <summary>
        /// Get the type of the map.
        /// </summary>
        public Mode Type { get { return mData.Metadata.Type; } }

        /// <summary>
        /// Get the author of the map.
        /// </summary>
        public string Author { get { return mData.Metadata.Author; } }

        /// <summary>
        /// Get a set of integers containing each allowable number of players.
        /// </summary>
        public HashSet<int> NumPlayers { get { return mData.Metadata.NumPlayers; } }

        /// <summary>
        /// Get the width of the map, in grid units.
        /// </summary>
        public int Width { get { return mData.Metadata.GridWidth; } }

        /// <summary>
        /// Get the height of the map, in grid units.
        /// </summary>
        public int Height { get { return mData.Metadata.GridHeight; } }

        /// <summary>
        /// Get the name of the tileset.
        /// </summary>
        public string Tileset { get { return mData.Metadata.Tileset; } }

        /// <summary>
        /// Get the current grid of open cells.  On the grid, true means
        /// the cell is open (i.e. an entity can occupy that cell), and false
        /// means the cell is closed.  Note that, just like Map.IsCellOpen,
        /// only static scenery is considered; the grid cannot tell you
        /// whether or not an entity is already occupying the cell.
        /// </summary>
        public bool[,] Grid { get { return mData.Grid; } }

        /// <summary>
        /// Get a list of the raw entity containers loaded with the map.  If you
        /// want to get the actual entity objects, use Map.GetEntities and
        /// Map.GetAllEntities instead.
        /// </summary>
        public List<RawEntity> RawEntities { get { return mData.Entities; } }


        /// <summary>
        /// Get and set the coordinate of the grid cell that should be in
        /// the center of the screen when the map is drawn.  Setting this
        /// will change the viewport of the map and will effect the return
        /// values of Map.GetPointFromCoordinates and Map.GetRectangleFromCoordinates.
        /// </summary>
        public Vector2 CenterCell
        {
            get { return mView.CenterCell; }
            set { mView.CenterCell = value; }
        }

        /// <summary>
        /// Get and set the tilemap with its associated texture and tile
        /// character to coordinate mappings.  This effects what the map looks
        /// like when it is drawn.
        /// </summary>
        public Tilemap Tilemap
        {
            get { return mView.Tilemap; }
            set { mView.Tilemap = value; }
        }

        /// <summary>
        /// Get and set the zoom of the map view.  The default zoom is
        /// Map.PixelsToUnitSquares.
        /// </summary>
        public float Zoom
        {
            get { return mView.Zoom; }
            set { mView.Zoom = value; }
        }

        #endregion


        #region Public Methods

        /// <summary>
        /// Construct a map with the provided map data.
        /// </summary>
        /// <param name="metadata">The metadata.</param>
        /// <param name="grid">The grid.</param>
        /// <param name="entities">The entities.</param>
        public Map(Metadata metadata, char[,] grid, char defaultTile,
            List<RawEntity> entities, Point[] playerPositions)
        {
            mData = new Model(metadata, grid, defaultTile, entities, playerPositions);
            mView = new View(mData);
        }


        /// <summary>
        /// Draw a representation of the map to the screen.
        /// </summary>
        /// <param name="spriteBatch">The jeewiz.</param>
        public void Draw(SpriteBatch spriteBatch)
        {
            mView.Draw(spriteBatch);
        }


        /// <summary>
        /// Get a point in screen-space from a coordinate in gridspace.
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        /// <returns>Transformed point.</returns>
        public Point GetPointFromCoordinates(float x, float y)
        {
            return mView.GetPointFromCoordinates(x, y);
        }

        /// <summary>
        /// Get a point in screen-space from a coordinate in gridspace.
        /// </summary>
        /// <param name="point">X,Y-coordinates.</param>
        /// <returns>Transformed point.</returns>
        public Point GetPointFromCoordinates(Vector2 point)
        {
            return mView.GetPointFromCoordinates(point.X, point.Y);
        }

        /// <summary>
        /// Get a rectangle in screen-space centered around a coordinate in gridspace.
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        /// <returns>Transformed rectangle.</returns>
        public Rectangle GetRectangleFromCoordinates(float x, float y)
        {
            return mView.GetRectangleFromCoordinates(x, y);
        }

        /// <summary>
        /// Get a rectangle in screen-space centered around a coordinate in gridspace.
        /// </summary>
        /// <param name="point">X,Y-coordinates.</param>
        /// <returns>Transformed rectangle.</returns>
        public Rectangle GetRectangleFromCoordinates(Vector2 point)
        {
            return mView.GetRectangleFromCoordinates(point.X, point.Y);
        }


        /// <summary>
        /// Determine whether or not a cell can be occupied by a game entity.
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        /// <returns>True if cell can be occupied, false otherwise.</returns>
        public bool IsCellOpen(int x, int y)
        {
            return mData.IsCellOpen(x, y);
        }

        /// <summary>
        /// Determine whether or not a cell can be occupied by a game entity.
        /// </summary>
        /// <param name="point">X,Y-coordinates.</param>
        /// <returns>True if cell can be occupied, false otherwise.</returns>
        public bool IsCellOpen(Point point)
        {
            return mData.IsCellOpen(point.X, point.Y);
        }


        /// <summary>
        /// Get the starting position of a player.
        /// </summary>
        /// <param name="playerNumber">The number of the player (i.e. 1-4).
        /// This number must be a valid player number.</param>
        /// <returns>The starting position of the player.</returns>
        public Point GetStartingPositionForPlayer(int playerNumber)
        {
            Debug.Assert(1 <= playerNumber && playerNumber <= NumPlayers.Max());
            return mData.PlayerPositions[playerNumber];
        }


        /// <summary>
        /// Get all the entities loaded from the map file.  Exceptions could be
        /// thrown if there are entities without associated classes.
        /// </summary>
        /// <param name="game">The game reference to be passed to entities.</param>
        /// <returns>List of entity objects loaded.</returns>
        public List<IEntity> GetAllEntities(Game game)
        {
            return mData.GetAllEntities(game);
        }

        /// <summary>
        /// Get the entities of a certain type loaded from the map file.  Exceptions
        /// could be thrown if there are entities without associated classes.
        /// </summary>
        /// <param name="game">The game reference to be passed to entities.</param>
        /// <typeparam name="T">Type of the entity you want a list of.</typeparam>
        /// <returns>List of entity objects loaded.</returns>
        public List<T> GetEntities<T>(Game game)
        {
            return mData.GetEntities<T>(game);
        }


        /// <summary>
        /// Set the tile of a cell.
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        /// <param name="tile">The character representing the tile.</param>
        public void SetCell(int x, int y, char tile)
        {
            mData.SetCell(x, y, tile);
        }

        /// <summary>
        /// Set the tile of a cell.
        /// </summary>
        /// <param name="point">X,Y-coordinates.</param>
        /// <param name="tile">The character representing the tile.</param>
        public void SetCell(Point point, char tile)
        {
            mData.SetCell(point.X, point.Y, tile);
        }


        /// <summary>
        /// Clear a cell to the default tile.
        /// </summary>
        /// <param name="x">X-coordinate.</param>
        /// <param name="y">Y-coordinate.</param>
        public void ClearCell(int x, int y)
        {
            mData.ClearCell(x, y);
        }

        /// <summary>
        /// Clear a cell to the default tile.
        /// </summary>
        /// <param name="point">X,Y-coordinates.</param>
        public void ClearCell(Point point)
        {
            mData.ClearCell(point.X, point.Y);
        }


        /// <summary>
        /// Reset the map to the state it was at right after loading.
        /// </summary>
        public void Reset()
        {
            mData.Reset();
            mView.Reset();
        }

        #endregion


        #region Private Types

        class Model
        {
            public Metadata Metadata { get { return mMetadata; } }
            public List<RawEntity> Entities { get { return mEntities; } }
            public Point[] PlayerPositions { get { return mPlayerPositions; } }
            public bool[,] Grid { get { return mBooleanGrid; } }


            public Model(Metadata metadata, char[,] grid, char defaultTile,
                List<RawEntity> entities, Point[] playerPositions)
            {
                Debug.Assert(metadata != null);
                Debug.Assert(grid != null);
                Debug.Assert(entities != null);
                Debug.Assert(metadata.GridWidth * metadata.GridHeight == grid.Length);

                mMetadata = metadata;
                mCleanGrid = grid;
                mDefaultTile = defaultTile;
                mEntities = entities;
                mPlayerPositions = playerPositions;

                Reset();

#if DEBUG
                Console.WriteLine("Loaded map {0} of type {1} written by {2}.",
                    metadata.Name,
                    metadata.Type,
                    metadata.Author);
#endif
            }


            public void Reset()
            {
                mGrid = (char[,])mCleanGrid.Clone();

                mBooleanGrid = new bool[mMetadata.GridWidth, mMetadata.GridHeight];
                for (int x = 0; x < mMetadata.GridWidth; x++)
                {
                    for (int y = 0; y < mMetadata.GridHeight; y++)
                    {
                        mBooleanGrid[x, y] = IsCellOpen(x, y);
                    }
                }
            }


            public bool IsCellOpen(int x, int y)
            {
                // TODO: Still need to define characters for types of scenery.
                if (IsOnMap(x, y)) return mGrid[x, y] == ' ';
                return false;
            }

            public void SetCell(int x, int y, char tile)
            {
                if (IsOnMap(x, y))
                {
                    mGrid[x, y] = tile;
                    mBooleanGrid[x, y] = IsCellOpen(x, y);
                }
            }

            public char GetCell(int x, int y)
            {
                if (IsOnMap(x, y)) return mGrid[x, y];
                return mDefaultTile;
            }

            public void ClearCell(int x, int y)
            {
                SetCell(x, y, mDefaultTile);
            }

            public bool IsOnMap(int x, int y)
            {
                return 0 <= x && x < Metadata.GridWidth && 0 <= y && y < Metadata.GridHeight;
            }


            public List<IEntity> GetAllEntities(Game game)
            {
                List<IEntity> list = new List<IEntity>();

                foreach (RawEntity raw in mEntities)
                {
                    if (raw.Attributes.ContainsKey("type"))
                    {
                        string typename = raw.Attributes["type"];

                        object[] args = new object[4];
                        args[0] = raw.Id;
                        args[1] = raw.Position;
                        args[2] = raw.Attributes;
                        args[3] = game;

                        try
                        {
                            IEntity entity = (IEntity)Activator.CreateInstance(System.Type.GetType("CarFire." + typename), args);
                            if (entity != null)
                            {
                                entity.LoadContent(game.ContentManager);
                                list.Add(entity);
                            }
                            else throw new Exception();
                        }
#pragma warning disable 0168
                        catch (System.Exception ex)
#pragma warning restore 0168
                        {
                            throw new Exception("Entity of type " + typename + " not loaded because an entity class can't be found.");
                        }
                    }
                    else
                    {
                        Console.WriteLine("Ignoring entity with identifier " + raw.Id + " since it has no type key.");
                    }
                }

                return list;
            }

            public List<T> GetEntities<T>(Game game)
            {
                System.Type type = typeof(T);
                List<T> list = new List<T>();

                string typename = typeof(T).Name;
                foreach (RawEntity raw in mEntities)
                {
                    if (raw.Attributes.ContainsKey("type") && typename == raw.Attributes["type"])
                    {
                        object[] args = new object[4];
                        args[0] = raw.Id;
                        args[1] = raw.Position;
                        args[2] = raw.Attributes;
                        args[3] = game;

                        T entity = (T)Activator.CreateInstance(type, args);
                        if (entity != null)
                        {
                            ((IEntity)entity).LoadContent(game.ContentManager);
                            list.Add(entity);
                        }
                        else throw new Exception("Entity of type " + typename + " not loaded because an entity class can't be found.");
                    }
                }

                return list;
            }


            Metadata mMetadata;
            char[,] mGrid;
            char[,] mCleanGrid;
            bool[,] mBooleanGrid;
            char mDefaultTile;
            List<RawEntity> mEntities;
            Point[] mPlayerPositions;
        }

        class View
        {
            public Vector2 CenterCell;
            public Tilemap Tilemap;
            public float Zoom;


            public View(Model data)
            {
                Debug.Assert(data != null);
                mData = data;

                Reset();
            }


            public void Reset()
            {
                CenterCell = Vector2.Zero;
                Zoom = PixelsToUnitSquares;
            }


            public void Draw(SpriteBatch spriteBatch)
            {
                if (Tilemap == null) throw new Exception("Cannot draw map without first setting the tilemap.");
                mViewport = spriteBatch.GraphicsDevice.Viewport;

                for (int y = 0; y < mData.Metadata.GridHeight; y++)
                {
                    for (int x = 0; x < mData.Metadata.GridWidth; x++)
                    {
                        Tilemap.Draw(spriteBatch, mData.GetCell(x, y), GetRectangleFromCoordinates(x, y));
                    }
                }
            }


            public Point GetPointFromCoordinates(float x, float y)
            {
                Matrix transform = GetTransformation(CenterCell);
                Vector2 point = Vector2.Transform(new Vector2(x, y), transform);

                return new Point((int)point.X, (int)point.Y);
            }

            public Rectangle GetRectangleFromCoordinates(float x, float y)
            {
                Matrix transform = GetTransformation(CenterCell);
                Vector2 point = Vector2.Transform(new Vector2(x, y), transform);
                
                return new Rectangle((int)Math.Round(point.X, 0), (int)Math.Round(point.Y, 0), (int)Math.Round(Zoom, 0), (int)Math.Round(Zoom, 0));
            }



            Matrix GetTransformation(Vector2 center)
            {
                float halfRatio = Zoom * 0.5f;
                Matrix transform = Matrix.CreateTranslation(-center.X, -center.Y, 0.0f);
                transform *= Matrix.CreateScale(Zoom);
                transform *= Matrix.CreateTranslation(mViewport.Width * 0.5f - halfRatio,
                    mViewport.Height * 0.5f - halfRatio, 0.0f);

                Vector2 topLeft = Vector2.Transform(new Vector2(0.0f, 0.0f), transform);
                topLeft.X = Math.Max(mViewport.X, topLeft.X);
                topLeft.Y = Math.Max(mViewport.Y, topLeft.Y);
                transform *= Matrix.CreateTranslation(-topLeft.X, -topLeft.Y, 0.0f);

                Vector2 bottomRight = Vector2.Transform(new Vector2((float)mData.Metadata.GridWidth,
                    (float)mData.Metadata.GridHeight), transform);
                float right = mViewport.X + mViewport.Width;
                float bottom = mViewport.Y + mViewport.Height;
                bottomRight.X = Math.Min(right, bottomRight.X) - right;
                bottomRight.Y = Math.Min(bottom, bottomRight.Y) - bottom;
                transform *= Matrix.CreateTranslation(-bottomRight.X, -bottomRight.Y, 0.0f);

                return transform;
            }


            Model mData;
            Viewport mViewport;
        }

        #endregion


        #region Private Variables

        Model mData;
        View mView;

        #endregion
    }
}
